from .automation import Automation, COLORS
from .exceptions import AIHeroException


class TagEntireImages(Automation):
    def get_tags(self):
        automation_definition = super().get_definition()
        ontology = automation_definition.get("ontology", [])
        return [c["name"] for c in ontology]

    def set_tags(self, tags):
        automation_definition = super().get_definition()
        ontology = []
        existing_in_ontology = []
        colors_to_select_from = [c for c in COLORS]

        for c in automation_definition.get("ontology", []):
            if c not in tags:
                continue  # Remove tags no longer needed
            ontology.append(c)
            existing_in_ontology.append(c["name"])
            colors_to_select_from.remove(c["color"])

        for t in tags:
            if t not in existing_in_ontology:
                ontology.append(
                    {
                        "type": "class",
                        "_id": t,
                        "name": t,
                        "color": colors_to_select_from.pop(),
                    }
                )

        super().update_ontology(ontology)
        return self.get_tags()

    def add(self, image_url, guid):
        if image_url is None or image_url.strip() == "":
            raise AIHeroException(
                "You need to provide the image to teach the automation with."
            )
        if guid is None or guid.strip() == "":
            raise AIHeroException(
                "You need to provide the guid to teach the automation with."
            )

        return super()._sync_job(
            {
                "type": "add_image",
                "row": {
                    "image": {"type": "url_pointer", "url": image_url},
                    "guid": guid,
                },
            }
        )

    def predict(self, image_url=None):
        if image_url is None or image_url.strip() == "":
            raise AIHeroException("image_url cannot be null or empty.")
        return super()._infer(
            "predict", {"image": {"type": "url_pointer", "url": image_url}}
        )



from .api import Api
from .automation import Automation, construct
from .workspace import Workspace


class Client:
    """Main sync client class to talk to AI Hero"""

    def __init__(self, server_url=None):
        self._api = Api(server_url=server_url)

    def ping(self):
        return self._api.get(
            self._api.ping_endpoint(),
            error_msg="Could not connect to the server {self._api.server_url}",
        )

    def get_automation(self, automation_id, api_key):
        automation = Automation(
            automation_id=automation_id,
            api=Api(api_key=api_key, server_url=self._api.server_url),
        )
        automation_definition = automation.get_definition()
        automation_type = automation_definition["type"]
        automation_api = Api(api_key=api_key, server_url=self._api.server_url)
        automation = construct(
            automation_type=automation_type,
            automation_id=automation_id,
            automation_api=automation_api,
        )
        return automation

    def request_auth_secret(self, email):
        path = f"{self._api.endpoint('users', 'auth_secret')}?email={email}"
        self._api.get(path)

    def get_workspace(self, auth_secret):
        workspace = Workspace(
            api=Api(auth_secret=auth_secret, server_url=self._api.server_url),
        )
        _ = workspace.get_automations()
        return workspace



from .automation import Automation
from .exceptions import AIHeroException


class DetectSentiment(Automation):
    def add(self, text, guid):
        if text is None or text.strip() == "":
            raise AIHeroException(
                "You need to provide the text to teach the automation with."
            )
        if guid is None or guid.strip() == "":
            raise AIHeroException(
                "You need to provide the guid to teach the automation with."
            )

        return super()._sync_job(
            {"type": "add_short_text", "row": {"text": text, "guid": guid}}
        )

    def predict(self, text):
        if text is None or text.strip() == "":
            raise AIHeroException("text cannot be null or empty.")
        return super()._infer("predict", {"text": text})



from .automation import Automation
from .exceptions import AIHeroException


class RecommendPeopleFromTwoGroupsToEachOther(Automation):
    def add_person_to_group_one(self, person, guid):
        if person is None or len(person) == 0:
            raise AIHeroException(
                "You need to provide the image to teach the automation with."
            )
        if guid is None or guid.strip() == "":
            raise AIHeroException(
                "You need to provide the guid to teach the automation with."
            )

        person["guid"] = guid
        return super()._sync_job({"type": "add_person_to_group_one", "row": person})

    def add_person_to_group_two(self, person, guid):
        if person is None or len(person) == 0:
            raise AIHeroException(
                "You need to provide the image to teach the automation with."
            )
        if guid is None or guid.strip() == "":
            raise AIHeroException(
                "You need to provide the guid to teach the automation with."
            )

        person["guid"] = guid
        return super()._sync_job({"type": "add_person_to_group_two", "row": person})

    def get_recommendations(self, person_id):
        if person_id is None or person_id.strip() == "":
            raise AIHeroException("person_id cannot be null or empty.")
        return super()._infer("get_recommendations", {"thing_id": person_id})



from aihero.exceptions import AIHeroException
from .automation import AUTOMATION_TYPES


class Workspace:
    """Abstraction for workspace operations"""

    def __init__(self, api):
        self._api = api

    def get_automations(self):
        return self._api.get(
            self._api.endpoint("workspace", "automations"),
            error_msg="Couldn't get the automations.",
            network_errors={
                403: "Could not connect to workspace. Please check the auth_secret."
            },
        )

    def create_automation(
        self, automation_type: str, name: str = None, description: str = None
    ):
        if automation_type not in AUTOMATION_TYPES:
            raise AIHeroException(
                f"Unknown/unsupported automation type {automation_type}"
            )
        obj = {"type": automation_type, "name": "auto2", "description": "Bar"}
        if name:
            obj["name"] = name
        if description:
            obj["description"] = description

        automation = self._api.post(
            self._api.endpoint("workspace", "automations"),
            obj=obj,
            error_msg="Couldn't create the automation api_key.",
            network_errors={
                403: "Could not create api key. Please check the auth_secret."
            },
        )
        return self.get_automation(automation["_id"])

    def get_automation(self, automation_id):
        api_keys = self._api.get(
            self._api.endpoint("workspace", "automations", automation_id, "api_keys"),
            error_msg="Couldn't get the automation api_key.",
            network_errors={
                403: "Could not get api key. Please check the auth_secret."
            },
        )
        if not api_keys:
            print(
                "Did not find an existing API key for the automation. Creating an API key."
            )
            self._api.post(
                self._api.endpoint(
                    "workspace", "automations", automation_id, "api_keys"
                ),
                obj={},
                error_msg="Couldn't create the automation api_key.",
                network_errors={
                    403: "Could not create api key. Please check the auth_secret."
                },
            )
            api_keys = self._api.get(
                self._api.endpoint(
                    "workspace", "automations", automation_id, "api_keys"
                ),
                error_msg="Couldn't get the automation api_key.",
                network_errors={
                    403: "Could not get api key. Please check the auth_secret."
                },
            )
        api_key = api_keys[0]["key"]
        from . import Client

        client = Client(server_url=self._api.server_url)
        return client.get_automation(automation_id=automation_id, api_key=api_key)



class AIHeroException(Exception):
    """Generic AI Hero exception class"""

    def __init__(self, message, status_code=None):
        if not message:
            message = type(self).__name__

        self.message = message
        if status_code:
            super().__init__(f"<Response [{status_code}]> {message}")
        else:
            super().__init__(f"{message}")



import httpx
from .exceptions import AIHeroException

PRODUCTION_URL = "https://api.aihero.studio"


class Api:
    """Abstraction for http operations"""

    def __init__(self, auth_secret=None, api_key=None, server_url=None):
        if auth_secret is None or auth_secret.strip() == "":
            self._auth_secret = None  # Only works for ping for now.
        else:
            self._auth_secret = auth_secret
            self._auth_secret_authorization = f"Bearer {self._auth_secret}"

        if api_key is None or api_key.strip() == "":
            self._api_key = None  # Only works for ping for now.
        else:
            self._api_key = api_key
            self._api_key_authorization = f"Bearer {self._api_key}"

        if server_url.endswith("/"):
            server_url = server_url[:-1]
        self.server_url = server_url or PRODUCTION_URL

    def _get_headers(self, path):
        headers = {
            "Content-Type": "application/json",
        }
        if "/workspace/" in path:
            # Prioritize workspace auth_key
            headers["Authorization"] = self._auth_secret_authorization
        elif "/automations/" in path:
            # api_key for automation
            headers["Authorization"] = self._api_key_authorization
        return headers

    def ping_endpoint(self):
        return f"{self.server_url}/ping"

    def endpoint(self, *args):
        return f"{self.server_url}/api/{'/'.join(args)}"

    def get(self, path, error_msg=None, network_errors=None):
        with httpx.Client() as client:
            response = client.get(path, headers=self._get_headers(path))
        try:
            response.raise_for_status()
        except httpx.HTTPStatusError as exc:
            if network_errors and exc.response.status_code in network_errors:
                raise AIHeroException(
                    f"{network_errors[exc.response.status_code]}",
                    status_code=exc.response.status_code,
                ) from exc
            elif exc.response.status_code:
                raise AIHeroException(
                    error_msg, status_code=exc.response.status_code
                ) from exc
            else:
                raise AIHeroException(error_msg) from exc
        return response.json()

    def post(self, path, obj, error_msg=None, network_errors=None):
        with httpx.Client() as client:
            response = client.post(path, json=obj, headers=self._get_headers(path))
        try:
            response.raise_for_status()
        except httpx.HTTPStatusError as exc:
            if network_errors and exc.response.status_code in network_errors:
                raise AIHeroException(
                    f"{network_errors[exc.response.status_code]}",
                    status_code=exc.response.status_code,
                ) from exc
            elif exc.response.status_code:
                raise AIHeroException(
                    error_msg, status_code=exc.response.status_code
                ) from exc
            else:
                raise AIHeroException(error_msg) from exc
        return response.json()



from .exceptions import AIHeroException
from time import sleep


COLORS = [
    "red",
    "orange",
    "amber",
    "yellow",
    "lime",
    "green",
    "emerald",
    "teal",
    "cyan",
    "sky",
    "blue",
    "indigo",
    "violet",
    "purple",
    "fuscia",
]

AUTOMATION_TYPES = [
    "detect_sentiment",
    "tag_short_text",
    "tag_entire_images",
    "recommend_items_to_people",
    "recommend_people_to_each_other",
    "recommend_people_from_two_groups_to_each_other",
]


def construct(automation_type, automation_id, automation_api):
    from .detect_sentiment import DetectSentiment
    from .tag_short_text import TagShortText
    from .tag_entire_images import TagEntireImages
    from .recommend_people_to_each_other import RecommendPeopleToEachOther
    from .recommend_people_from_two_groups_to_each_other import (
        RecommendPeopleFromTwoGroupsToEachOther,
    )
    from .recommend_items_to_people import RecommendItemsToPeople

    AUTOMATION_CLASSES = {
        "detect_sentiment": DetectSentiment,
        "tag_short_text": TagShortText,
        "tag_entire_images": TagEntireImages,
        "recommend_items_to_people": RecommendItemsToPeople,
        "recommend_people_to_each_other": RecommendPeopleToEachOther,
        "recommend_people_from_two_groups_to_each_other": RecommendPeopleFromTwoGroupsToEachOther,
    }

    if automation_type not in AUTOMATION_CLASSES:
        raise AIHeroException(
            f"Unsupported automation type for the API: {automation_type}"
        )
    return AUTOMATION_CLASSES[automation_type](
        automation_id=automation_id, api=automation_api
    )


class Automation:
    """Abstraction for automation operations"""

    def __init__(self, automation_id, api):
        if automation_id is None or automation_id.strip() == "":
            raise AIHeroException("automation_id cannot be null or empty.")
        self._api = api
        self._automation_id = automation_id

    def get_definition(self):
        return self._api.get(
            self._api.endpoint("automations", self._automation_id, "definition"),
            error_msg="Couldn't get the definition.",
            network_errors={
                403: f"Could not connect to automation {self._automation_id}. Please check the API key."
            },
        )

    def update_ontology(self, ontology):
        return self._api.post(
            self._api.endpoint(
                "automations", self._automation_id, "definition", "ontology"
            ),
            ontology,
            error_msg="Unknown error in updating ontology.",
            network_errors={
                400: "Please check the ontology to be updated.",
                403: f"Could not connect to automation {self._automation_id}. Please check the API key.",
            },
        )

    def _sync_job(self, job):
        job = self._api.post(
            self._api.endpoint("automations", self._automation_id, "jobs"),
            job,
            error_msg="Unknown error in job.",
            network_errors={
                400: "Please check the job parameters.",
                403: f"Could not connect to automation {self._automation_id}. Please check the API key.",
            },
        )
        while True:
            sleep(0.1)
            job = self._api.get(
                self._api.endpoint(
                    "automations", self._automation_id, "jobs", job["_id"]
                ),
                error_msg="Unknown error in job.",
                network_errors={
                    400: "Please check the request.",
                    403: f"Could not connect to automation {self._automation_id}. Please check the API key.",
                },
            )
            if "state" not in job:
                continue
            state = job["state"]
            if state == "created":
                pass
            elif state not in ["done", "continue", "error"]:
                pass
            elif state == "error":
                raise AIHeroException("Error while uploading data")
            elif state in ["done"]:
                break

    def _async_job(self, job):
        job = self._api.post(
            self._api.endpoint("automations", self._automation_id, "jobs"),
            job,
            error_msg="Unknown error in job.",
            network_errors={
                400: "Please check the job parameters.",
                403: f"Could not connect to automation {self._automation_id}. Please check the API key.",
            },
        )
        return job

    def _infer(self, task, obj):
        return self._api.post(
            self._api.endpoint("automations", self._automation_id, "inferences", task),
            obj,
            error_msg="Unknown error in {task}.",
            network_errors={
                400: "Please check the data uploaded.",
                403: f"Could not connect to automation {self._automation_id}. Please check the API key.",
            },
        )

    def relearn(self):
        return self._async_job({"type": "relearn"})

    def get_job(self, job):
        return self._api.get(
            self._api.endpoint("automations", self._automation_id, "jobs", job["_id"]),
            error_msg="Unknown error in job.",
            network_errors={
                400: "Please check the request.",
                403: f"Could not connect to automation {self._automation_id}. Please check the API key.",
            },
        )



from .automation import Automation
from .exceptions import AIHeroException


class RecommendItemsToPeople(Automation):
    def add_item(self, item, guid):
        if item is None or len(item) == 0:
            raise AIHeroException(
                "You need to provide the image to teach the automation with."
            )
        if guid is None or guid.strip() == "":
            raise AIHeroException(
                "You need to provide the guid to teach the automation with."
            )

        item["guid"] = guid
        return super()._sync_job({"type": "add_item", "row": item})

    def get_recommendations(self, item_id):
        if item_id is None or item_id.strip() == "":
            raise AIHeroException("item_id cannot be null or empty.")
        return super()._infer("get_recommendations", {"thing_id": item_id})
